import pytest
import torch

def test_grouped_ff_one_group():
    from x_mlps_pytorch.grouped_ff import GroupedFeedforwards

    ff = GroupedFeedforwards(256, 4, dim_in = 128, dim_out = 128, squeeze_if_one_group = True)

    x = torch.randn(7, 3, 128)

    assert ff(x).shape == x.shape


def test_grouped_ff_one_group():
    from x_mlps_pytorch.grouped_ff import GroupedFeedforwards

    ff = GroupedFeedforwards(256, 4, dim_in = 128, dim_out = 128, squeeze_if_one_group = True, groups = 2)

    x = torch.randn(7, 3, 128)

    assert ff(x).shape == (7, 3, 2, 128)



import pytest
import torch

def test_ff():
    from x_mlps_pytorch.ff import Feedforwards

    ff = Feedforwards(256, 4, dim_in = 128, dim_out = 128)

    x = torch.randn(7, 3, 128)

    assert ff(x).shape == x.shape

@pytest.mark.parametrize('preserve_magnitude', (False, True))
def test_nff(
    preserve_magnitude
):
    from x_mlps_pytorch.nff import nFeedforwards, norm_weights_

    ff = nFeedforwards(256, 4, input_preserve_magnitude = preserve_magnitude)

    x = torch.randn(7, 3, 256)

    assert ff(x).shape == x.shape

    norm_weights_(ff)

@pytest.mark.parametrize('input_has_seq_dim', (False, True))
@pytest.mark.parametrize('latent_mlp', (False, True))
def test_ff_with_latent(
    input_has_seq_dim,
    latent_mlp
):
    from x_mlps_pytorch.ff_with_latent import LatentConditionedFeedforwards

    ff = LatentConditionedFeedforwards(256, 4, dim_in = 128, dim_out = 128, dim_latent = 33, latent_mlp = latent_mlp)

    x = torch.randn(7, 3, 128) if input_has_seq_dim else torch.randn(7, 128)
    latent = torch.randn(7, 33)

    assert ff(x, latent = latent).shape == x.shape



import torch
from torch import nn, cat
from torch.nn import Module, ModuleList

# functions

def exists(v):
    return v is not None

# main class

class Feedforwards(Module):

    def __init__(
        self,
        dim,
        depth,
        *,
        dim_in = None,
        dim_out = None,
        activation = nn.GELU(),
        bias = True,
        expansion_factor = 4.,
        norm_after_activation = False,
        final_norm = False
    ):
        super().__init__()

        layers = []

        dim_hidden = int(dim * expansion_factor)

        # layers

        for _ in range(depth):

            layer = nn.Sequential(
                nn.RMSNorm(dim),
                nn.Linear(dim, dim_hidden, bias = bias),
                activation,
                nn.RMSNorm(dim) if norm_after_activation else nn.Identity(),
                nn.Linear(dim_hidden, dim, bias = bias)
            )

            layers.append(layer)

        self.layers = ModuleList(layers)

        # maybe final norm

        self.norm = nn.RMSNorm(dim) if final_norm else nn.Identity()

        # proj in and out

        self.proj_in = nn.Linear(dim_in, dim) if exists(dim_in) else nn.Identity()
        self.proj_out = nn.Linear(dim, dim_out) if exists(dim_out) else nn.Identity()

    def forward(
        self,
        x
    ):

        if isinstance(x, (list, tuple)):
            x = cat(x, dim = -1)

        x = self.proj_in(x)

        for layer in self.layers:
            x = layer(x) + x

        x = self.norm(x)

        return self.proj_out(x)



from copy import deepcopy

import torch
from torch import Tensor, tensor, is_tensor

from torch.nn import Module, ParameterList
from torch.func import vmap, functional_call

from einops import einsum

# functions

def exists(v):
    return v is not None

def default(v, d):
    return v if exists(v) else d

# main class

class Ensemble(Module):
    def __init__(
        self,
        net: Module,
        ensemble_size,
        init_std_dev = 2e-2
    ):
        super().__init__()
        self.net = net
        self.ensemble_size = ensemble_size

        params = dict(net.named_parameters())
        device = next(iter(params.values())).device

        ensemble_params = {name: (torch.randn((ensemble_size, *param.shape), device = device) * init_std_dev).requires_grad_() for name, param in params.items()}

        self.param_names = list(ensemble_params.keys())
        self.param_values = ParameterList(list(ensemble_params.values()))

        def _forward(params, args, kwargs):
            return functional_call(net, params, args, kwargs)

        self._forward = _forward
        self.ensemble_forward = vmap(_forward, in_dims = (0, None, None))
        self.each_batch_sample_forward = vmap(_forward, in_dims = (0, 0, 0))

    @property
    def ensemble_params(self):
        return dict(zip(self.param_names, self.param_values))

    def parameters(self):
        return iter(self.ensemble_params.values())

    @property
    def device(self):
        return next(self.parameters()).device

    def pick_params(
        self,
        indices: int | Tensor | list[int] | None = None
    ):
        params = self.ensemble_params

        if not exists(indices):
            return params

        # converts list[int] to Int

        if isinstance(indices, list):
            indices = tensor(indices, device = self.device)

        # some validation

        if is_tensor(indices):
            assert (indices < self.ensemble_size).all()
        elif isinstance(indices, int):
            assert 0 <= indices < self.ensemble_size

        params = {key: param[indices] for key, param in params.items()}
        return params

    def get_one(
        self,
        indices: int | Tensor | list[int],
        weights: list[float] = None
    ):

        params = self.pick_params(indices)

        needs_combine = not isinstance(indices, int)

        if needs_combine:
            num_nets = len(indices)

            weights = default(weights, [1. / num_nets] * num_nets)
            weights = tensor(weights, device = self.device)

            params = {key: einsum(param, weights, 'n ..., n -> ...') for key, param in params.items()}

        out = deepcopy(self.net)
        out.load_state_dict(params)

        return out

    def forward_one(
        self,
        *args,
        id: int,
        **kwargs
    ):
        params = {key: param[id] for key, param in self.ensemble_params.items()}
        return self._forward(dict(params), args, kwargs)

    def forward(
        self,
        *args,
        ids = None,
        each_batch_sample = False,
        **kwargs,
    ):

        params = self.pick_params(ids)

        fn = self.ensemble_forward if not each_batch_sample else self.each_batch_sample_forward

        return fn(dict(params), args, kwargs)



from functools import partial

import torch
from torch.nn import Module, ReLU

# relu squared with optional signing

class ReluSquared(Module):
    def __init__(self, signed = False):
        super().__init__()
        self.signed = signed

    def forward(self, x):
        out = x.relu().square()

        if not self.signed:
            return out

        return out * x.sign()

# sugar-(bsilu | nelu)

class BSiLU(Module):
    # eq (7) in paper

    def __init__(self, alpha = 1.67):
        super().__init__()
        self.alpha = alpha

    def forward(self, x):
        α = self.alpha
        return (x + α) * x.sigmoid() - α / 2

class NeLU(Module):
    def __init__(self, alpha = 0.05):
        super().__init__()
        self.alpha = alpha

    def forward(self, x):
        α = self.alpha
        return -α / (1. + x.square())

class StraightThrough(Module):
    def __init__(
        self,
        forward_fn: Module,
        backward_fn: Module
    ):
        super().__init__()
        self.forward_fn = forward_fn
        self.backward_fn = backward_fn

    def forward(self, x):
        hard = self.forward_fn(x)

        if not x.requires_grad:
            return hard

        soft = self.backward_fn(x)

        # straight-through during training

        return soft + (hard - soft).detach()

class Sugar(Module):
    def __init__(
        self,
        forward_fn: Module,
        backward_fn: Module
    ):
        super().__init__()
        self.forward_fn = forward_fn
        self.backward_fn = backward_fn

    def forward(self, x):
        forward_out = self.forward_fn(x)

        if not x.requires_grad:
            return forward_out

        backward_out = self.backward_fn(x)

        # only neg region for backward function gradients

        soft = torch.where(x > 0, forward_out, backward_out)
        
        # straight-through during training

        return soft + (forward_out - soft).detach()

# the one that beat gelu in transformer setting for me

def ReluNelu(alpha = 0.05):
    return Sugar(ReLU(), NeLU(alpha))
